/*\

Copyright 2017 The University of North Carolina at Chapel Hill.
All Rights Reserved.

Permission to use, copy, modify and distribute this software and its
documentation for educational, research and non-profit purposes, without
fee, and without a written agreement is hereby granted, provided that the
above copyright notice and the following three paragraphs appear in all
copies.

IN NO EVENT SHALL THE UNIVERSITY OF NORTH CAROLINA AT CHAPEL HILL BE
LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR
CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS, ARISING OUT OF THE
USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY
OF NORTH CAROLINA HAVE BEEN ADVISED OF THE POSSIBILITY OF SUCH
DAMAGES.

THE UNIVERSITY OF NORTH CAROLINA SPECIFICALLY DISCLAIM ANY
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE
PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
NORTH CAROLINA HAS NO OBLIGATIONS TO PROVIDE MAINTENANCE, SUPPORT,
UPDATES, ENHANCEMENTS, OR MODIFICATIONS.

The authors may be contacted via:

US Mail: Real-Time Systems Group at UNC
Department of Computer Science
Sitterson Hall
University of N. Carolina
Chapel Hill, NC 27599-3175

EMail: nemitz@cs.unc.edu; tamert@cs.unc.edu; anderson@cs.unc.edu

**/

#ifndef FAST_RWRNLP_H
#define FAST_RWRNLP_H

#include <stdbool.h>
#include <stdint.h>

#include "mem.h"
#include "nesting.h"
#include "pft.h"
#include "rnlp.h"
#include "rwrnlp_star.h"
#include "ticket.h"

#define NR_CPUS 64
#define NR_RESOURCES 64

typedef struct fast_rwrnlp_struct {
    // Use the RW-RNLP* struct to handle everything
    rwrnlp_star_t rwrnlp_star;

    // One ticket lock per resource, to service nn-writes before
    // they get to the s_read_lock call
    ticket_lock_t ticketlocks[NR_RESOURCES];

    // One RNLP to service n-writes before they get to the
    // n_read_lock call
    rnlp_lock_t rnlp;

    // Store the RNLP requests (only for n-writes) per processor
    rnlp_node_t rnlpnodes[NR_CPUS];

    // Use a spinning ticket lock to control access to write structures
    ticket_lock_t *nested;
} fast_rwrnlp_t;

void fast_rwrnlp_init(fast_rwrnlp_t *lock);

// Reads (pass-through to RW-RNLP*)
void nn_read_lock(fast_rwrnlp_t *lock, int resource);
void nn_read_unlock(fast_rwrnlp_t *lock, int resource);
void n_read_lock(fast_rwrnlp_t *lock, resource_mask_t resources, int processor);
void n_read_unlock(fast_rwrnlp_t *lock, resource_mask_t resources, int processor);

// Non-Nested Writes (first go through ticket-lock)
void nn_write_lock(fast_rwrnlp_t *lock, int resource);
void nn_write_unlock(fast_rwrnlp_t *lock, int resource);

// Nested Writes (first go through RNLP)
void n_write_lock(fast_rwrnlp_t *lock, resource_mask_t resources, int processor);
void n_write_unlock(fast_rwrnlp_t *lock, resource_mask_t resources, int processor);

#endif // FAST_RWRNLP_H
