/*\

Copyright 2017 The University of North Carolina at Chapel Hill.
All Rights Reserved.

Permission to use, copy, modify and distribute this software and its
documentation for educational, research and non-profit purposes, without
fee, and without a written agreement is hereby granted, provided that the
above copyright notice and the following three paragraphs appear in all
copies.

IN NO EVENT SHALL THE UNIVERSITY OF NORTH CAROLINA AT CHAPEL HILL BE
LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR
CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS, ARISING OUT OF THE
USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY
OF NORTH CAROLINA HAVE BEEN ADVISED OF THE POSSIBILITY OF SUCH
DAMAGES.

THE UNIVERSITY OF NORTH CAROLINA SPECIFICALLY DISCLAIM ANY
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE
PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
NORTH CAROLINA HAS NO OBLIGATIONS TO PROVIDE MAINTENANCE, SUPPORT,
UPDATES, ENHANCEMENTS, OR MODIFICATIONS.

The authors may be contacted via:

US Mail: Real-Time Systems Group at UNC
Department of Computer Science
Sitterson Hall
University of N. Carolina
Chapel Hill, NC 27599-3175

EMail: nemitz@cs.unc.edu; tamert@cs.unc.edu; anderson@cs.unc.edu

**/

#include <stdio.h>
#include <stdlib.h>

#include "rwrnlp_star.h"

void rwrnlp_star_init(rwrnlp_star_t *lock)
{
    int i;

    for (i = 0; i < NR_RESOURCES; i++)
    {
        pft_lock_init(&lock->stuff[i]);
    }

    lock->nested = (pft_lock_t *)malloc(sizeof(pft_lock_t));
    pft_lock_init(lock->nested);
}

void rwrnlp_star_nn_read_lock(rwrnlp_star_t *lock, unsigned int resource)
{
    unsigned int w;

    // Second pass: now it's entitled; actually increment rin
    w = __sync_fetch_and_add(&lock->stuff[resource].rin, RINC) & WBITS;

    // Wait
    while ((w != 0) && (w == (lock->stuff[resource].rin & WBITS)))
    {
        cpu_relax();
    }
}

void rwrnlp_star_nn_read_unlock(rwrnlp_star_t *lock, unsigned int resource)
{
    __sync_fetch_and_add(&lock->stuff[resource].rout, RINC);
}

void rwrnlp_star_nn_write_lock(rwrnlp_star_t *lock, unsigned int resource)
{
    unsigned int w, rticket, wticket;

    // Wait for my turn
    wticket = __sync_fetch_and_add(&lock->stuff[resource].win, 1);

    while (wticket != lock->stuff[resource].wout)
    {
        cpu_relax();
    }

    // Set the write-bits of rin to indicate this writer is here
    w = PRES | (wticket & PHID);
    rticket = __sync_fetch_and_add(&lock->stuff[resource].rin, w);

    while (rticket != lock->stuff[resource].rout)
    {
        cpu_relax();
    }
}

void rwrnlp_star_nn_write_unlock(rwrnlp_star_t *lock, unsigned int resource)
{
    unsigned int andoperand;

    andoperand = -256;

    __sync_fetch_and_and(&lock->stuff[resource].rin, andoperand);
    lock->stuff[resource].wout++;
}

void rwrnlp_star_read_lock(rwrnlp_star_t *lock, unsigned long resources, int processor)
{
    unsigned int w[NR_RESOURCES];
    int r;
    unsigned long one, tmp;

    one = 1;

    // First pass, don't increment rin
    tmp = resources;
    r = ffsl(tmp);
    pft_write_lock(lock->nested);
    while (r != 0)
    {
        r = r - 1;

        w[r] = __sync_fetch_and_add(&lock->stuff[r].rin, 0) & WBITS;

        tmp &= ~(one << r);
        r = ffsl(tmp);
    }
    pft_write_unlock(lock->nested);

    // Wait
    tmp = resources;
    r = ffsl(tmp);
    while (r != 0)
    {
        r = r - 1;

        while ((w[r] != 0) && (w[r] == (lock->stuff[r].rin & WBITS)))
        {
            cpu_relax();
        }

        tmp &= ~(one << r);
        r = ffsl(tmp);
    }

    // Second pass, now it's entitled; actually increment rin
    tmp = resources;
    r = ffsl(tmp);
    pft_write_lock(lock->nested);
    while (r != 0)
    {
        r = r - 1;

        w[r] = __sync_fetch_and_add(&lock->stuff[r].rin, RINC) & WBITS;

        tmp &= ~(one << r);
        r = ffsl(tmp);
    }
    pft_write_unlock(lock->nested);

    // Wait
    tmp = resources;
    r = ffsl(tmp);
    while (r != 0)
    {
        r = r - 1;

        while ((w[r] != 0) && (w[r] == (lock->stuff[r].rin & WBITS)))
        {
            cpu_relax();
        }

        tmp &= ~(one << r);
        r = ffsl(tmp);
    }
}

void rwrnlp_star_read_unlock(rwrnlp_star_t *lock, unsigned long resources, int processor)
{
    int r;
    unsigned long one, tmp;

    one = 1;

    tmp = resources;
    r = ffsl(tmp);
    while (r != 0)
    {
        r = r - 1;

        __sync_fetch_and_add(&lock->stuff[r].rout, RINC);

        tmp &= ~(one << r);
        r = ffsl(tmp);
    }
}

void rwrnlp_star_write_lock(rwrnlp_star_t *lock, unsigned long resources, int processor)
{
    unsigned int w[NR_RESOURCES], rticket[NR_RESOURCES], wticket[NR_RESOURCES];
    int r;
    unsigned long one, tmp;

    one = 1;

    // Wait for my turn
    tmp = resources;
    r = ffsl(tmp);
    while (r != 0)
    {
        r = r - 1;

        wticket[r] = __sync_fetch_and_add(&lock->stuff[r].win, 1);

        tmp &= ~(one << r);
        r = ffsl(tmp);
    }

    tmp = resources;
    r = ffsl(tmp);
    while (r != 0)
    {
        r = r - 1;

        while (wticket[r] != lock->stuff[r].wout)
        {
            cpu_relax();
        }

        tmp &= ~(one << r);
        r = ffsl(tmp);
    }

    // Set the write-bits of rin to indicate this writer is here
    tmp = resources;
    r = ffsl(tmp);
    pft_read_lock(lock->nested);
    while (r != 0)
    {
        r = r - 1;

        w[r] = PRES | (wticket[r] & PHID);
        rticket[r] = __sync_fetch_and_add(&lock->stuff[r].rin, w[r]);

        tmp &= ~(one << r);
        r = ffsl(tmp);
    }
    pft_read_unlock(lock->nested);

    tmp = resources;
    r = ffsl(tmp);
    while (r != 0)
    {
        r = r - 1;

        while (rticket[r] != lock->stuff[r].rout)
        {
            cpu_relax();
        }

        tmp &= ~(one << r);
        r = ffsl(tmp);
    }
}

void rwrnlp_star_write_unlock(rwrnlp_star_t *lock, unsigned long resources, int processor)
{
    unsigned int andoperand;
    int r;
    unsigned long one, tmp;

    one = 1;
    andoperand = -256;

    tmp = resources;
    r = ffsl(tmp);
    while (r != 0)
    {
        r = r - 1;

        __sync_fetch_and_and(&lock->stuff[r].rin, andoperand);
        lock->stuff[r].wout++;

        tmp &= ~(one << r);
        r = ffsl(tmp);
    }
}
